@@ -1,9 +1,12 @@
 'use strict';
 
 const {
-  Map,
+  NumberParseInt,
   ObjectDefineProperty,
+  ObjectDefineProperties,
+  SafeMap,
   SafeWeakMap,
+  StringPrototypeStartsWith,
 } = primordials;
 
 const {
@@ -29,14 +32,6 @@ function prepareMainThreadExecution(expandArgv1 = false) {
       setupCoverageHooks(process.env.NODE_V8_COVERAGE);
   }
 
-  // If source-map support has been enabled, we substitute in a new
-  // prepareStackTrace method, replacing the default in errors.js.
-  if (getOptionValue('--enable-source-maps')) {
-    const { prepareStackTrace } =
-      require('internal/source_map/prepare_stack_trace');
-    const { setPrepareStackTraceCallback } = internalBinding('errors');
-    setPrepareStackTraceCallback(prepareStackTrace);
-  }
 
   setupDebugEnv();
 
@@ -64,6 +59,7 @@ function prepareMainThreadExecution(expandArgv1 = false) {
   // (including preload modules).
   initializeClusterIPC();
 
+  initializeAbortController();
   initializeDeprecations();
   initializeWASI();
   initializeCJSLoader();
@@ -89,10 +85,13 @@ function patchProcessObject(expandArgv1) {
   });
   process.argv[0] = process.execPath;
 
-  if (expandArgv1 && process.argv[1] && !process.argv[1].startsWith('-')) {
+  if (expandArgv1 && process.argv[1] &&
+      !StringPrototypeStartsWith(process.argv[1], '-')) {
     // Expand process.argv[1] into a full path.
     const path = require('path');
-    process.argv[1] = path.resolve(process.argv[1]);
+    try {
+      process.argv[1] = path.resolve(process.argv[1]);
+    } catch {}
   }
 
   // TODO(joyeecheung): most of these should be deprecated and removed,
@@ -225,9 +224,9 @@ function setupInspectorHooks() {
   }
 }
 
-// In general deprecations are intialized wherever the APIs are implemented,
+// In general deprecations are initialized wherever the APIs are implemented,
 // this is used to deprecate APIs implemented in C++ where the deprecation
-// utitlities are not easily accessible.
+// utilities are not easily accessible.
 function initializeDeprecations() {
   const { deprecate } = require('internal/util');
   const pendingDeprecation = getOptionValue('--pending-deprecation');
@@ -315,11 +314,35 @@ function initializeDeprecations() {
   });
 }
 
+function initializeAbortController() {
+  const abortController = getOptionValue('--experimental-abortcontroller');
+  if (abortController) {
+    const {
+      AbortController,
+      AbortSignal
+    } = require('internal/abort_controller');
+    ObjectDefineProperties(global, {
+      AbortController: {
+        writable: true,
+        enumerable: false,
+        configurable: true,
+        value: AbortController
+      },
+      AbortSignal: {
+        writable: true,
+        enumerable: false,
+        configurable: true,
+        value: AbortSignal
+      }
+    });
+  }
+}
+
 function setupChildProcessIpcChannel() {
   if (process.env.NODE_CHANNEL_FD) {
     const assert = require('internal/assert');
 
-    const fd = parseInt(process.env.NODE_CHANNEL_FD, 10);
+    const fd = NumberParseInt(process.env.NODE_CHANNEL_FD, 10);
     assert(fd >= 0);
 
     // Make sure it's not accidentally inherited by child processes.
@@ -348,7 +371,7 @@ function initializePolicy() {
   if (experimentalPolicy) {
     process.emitWarning('Policies are experimental.',
                         'ExperimentalWarning');
-    const { pathToFileURL, URL } = require('url');
+    const { pathToFileURL, URL } = require('internal/url');
     // URL here as it is slightly different parsing
     // no bare specifiers for now
     let manifestURL;
@@ -365,7 +388,7 @@ function initializePolicy() {
     if (experimentalPolicyIntegrity) {
       const SRI = require('internal/policy/sri');
       const { createHash, timingSafeEqual } = require('crypto');
-      const realIntegrities = new Map();
+      const realIntegrities = new SafeMap();
       const integrityEntries = SRI.parse(experimentalPolicyIntegrity);
       let foundMatch = false;
       for (let i = 0; i < integrityEntries.length; i++) {
@@ -451,6 +474,7 @@ module.exports = {
   setupWarningHandler,
   setupDebugEnv,
   prepareMainThreadExecution,
+  initializeAbortController,
   initializeDeprecations,
   initializeESMLoader,
   initializeFrozenIntrinsics,
