@@ -1,9 +1,16 @@
 'use strict';
 
 const {
+  ArrayPrototypeForEach,
+  ArrayPrototypeJoin,
   ObjectDefineProperty,
   ObjectPrototypeHasOwnProperty,
   SafeMap,
+  SafeSet,
+  StringPrototypeCharCodeAt,
+  StringPrototypeIncludes,
+  StringPrototypeSlice,
+  StringPrototypeStartsWith,
 } = primordials;
 const {
   ERR_MANIFEST_DEPENDENCY_MISSING,
@@ -13,13 +20,18 @@ const { NativeModule } = require('internal/bootstrap/loaders');
 
 const { validateString } = require('internal/validators');
 const path = require('path');
-const { pathToFileURL, fileURLToPath } = require('internal/url');
-const { URL } = require('url');
+const { pathToFileURL, fileURLToPath, URL } = require('internal/url');
+
+const { getOptionValue } = require('internal/options');
+const userConditions = getOptionValue('--conditions');
 
 let debug = require('internal/util/debuglog').debuglog('module', (fn) => {
   debug = fn;
 });
 
+// TODO: Use this set when resolving pkg#exports conditions in loader.js.
+const cjsConditions = new SafeSet(['require', 'node', ...userConditions]);
+
 function loadNativeModule(filename, request) {
   const mod = NativeModule.map.get(filename);
   if (mod) {
@@ -38,11 +50,12 @@ function makeRequireFunction(mod, redirects) {
 
   let require;
   if (redirects) {
-    const { resolve, reaction } = redirects;
     const id = mod.filename || mod.id;
-    require = function require(path) {
+    const conditions = cjsConditions;
+    const { resolve, reaction } = redirects;
+    require = function require(specifier) {
       let missing = true;
-      const destination = resolve(path);
+      const destination = resolve(specifier, conditions);
       if (destination === true) {
         missing = false;
       } else if (destination) {
@@ -66,9 +79,13 @@ function makeRequireFunction(mod, redirects) {
         }
       }
       if (missing) {
-        reaction(new ERR_MANIFEST_DEPENDENCY_MISSING(id, path));
+        reaction(new ERR_MANIFEST_DEPENDENCY_MISSING(
+          id,
+          specifier,
+          ArrayPrototypeJoin([...conditions], ', ')
+        ));
       }
-      return mod.require(path);
+      return mod.require(specifier);
     };
   } else {
     require = function require(path) {
@@ -106,8 +123,8 @@ function makeRequireFunction(mod, redirects) {
  * translates it to FEFF, the UTF-16 BOM.
  */
 function stripBOM(content) {
-  if (content.charCodeAt(0) === 0xFEFF) {
-    content = content.slice(1);
+  if (StringPrototypeCharCodeAt(content) === 0xFEFF) {
+    content = StringPrototypeSlice(content, 1);
   }
   return content;
 }
@@ -115,11 +132,11 @@ function stripBOM(content) {
 function addBuiltinLibsToObject(object) {
   // Make built-in modules available directly (loaded lazily).
   const { builtinModules } = require('internal/modules/cjs/loader').Module;
-  builtinModules.forEach((name) => {
+  ArrayPrototypeForEach(builtinModules, (name) => {
     // Neither add underscored modules, nor ones that contain slashes (e.g.,
     // 'fs/promises') or ones that are already defined.
-    if (name.startsWith('_') ||
-        name.includes('/') ||
+    if (StringPrototypeStartsWith(name, '_') ||
+        StringPrototypeIncludes(name, '/') ||
         ObjectPrototypeHasOwnProperty(object, name)) {
       return;
     }
@@ -168,6 +185,7 @@ function normalizeReferrerURL(referrer) {
 
 module.exports = {
   addBuiltinLibsToObject,
+  cjsConditions,
   loadNativeModule,
   makeRequireFunction,
   normalizeReferrerURL,
