@@ -3,6 +3,8 @@
 const {
   ArrayFrom,
   ArrayIsArray,
+  Error,
+  Map,
   ObjectCreate,
   ObjectDefineProperties,
   ObjectDefineProperty,
@@ -12,6 +14,7 @@ const {
   ObjectSetPrototypeOf,
   Promise,
   ReflectConstruct,
+  Set,
   Symbol,
   SymbolFor,
 } = primordials;
@@ -41,6 +44,13 @@ const experimentalWarnings = new Set();
 
 const colorRegExp = /\u001b\[\d\d?m/g; // eslint-disable-line no-control-regex
 
+let uvBinding;
+
+function lazyUv() {
+  uvBinding = uvBinding ?? internalBinding('uv');
+  return uvBinding;
+}
+
 function removeColors(str) {
   return str.replace(colorRegExp, '');
 }
@@ -268,7 +278,11 @@ function getSystemErrorName(err) {
   return entry ? entry[0] : `Unknown system error ${err}`;
 }
 
-const kCustomPromisifiedSymbol = Symbol('util.promisify.custom');
+function getSystemErrorMap() {
+  return lazyUv().getErrorMap();
+}
+
+const kCustomPromisifiedSymbol = SymbolFor('nodejs.util.promisify.custom');
 const kCustomPromisifyArgsSymbol = Symbol('customPromisifyArgs');
 
 function promisify(original) {
@@ -411,6 +425,7 @@ module.exports = {
   emitExperimentalWarning,
   filterDuplicateStrings,
   getConstructorOf,
+  getSystemErrorMap,
   getSystemErrorName,
   isError,
   isInsideNodeModules,
