@@ -23,7 +23,23 @@ const fixtureA = fixtures.path('printA.js');
 const fixtureB = fixtures.path('printB.js');
 const fixtureC = fixtures.path('printC.js');
 const fixtureD = fixtures.path('define-global.js');
+const fixtureE = fixtures.path('intrinsic-mutation.js');
+const fixtureF = fixtures.path('print-intrinsic-mutation-name.js');
+const fixtureG = fixtures.path('worker-from-argv.js');
 const fixtureThrows = fixtures.path('throws_error4.js');
+const fixtureIsPreloading = fixtures.path('ispreloading.js');
+
+// Assert that module.isPreloading is false here
+assert(!module.isPreloading);
+
+// Test that module.isPreloading is set in preloaded module
+// Test preloading a single module works
+childProcess.exec(
+  `"${nodeBinary}" ${preloadOption([fixtureIsPreloading])} "${fixtureB}"`,
+  function(err, stdout, stderr) {
+    assert.ifError(err);
+    assert.strictEqual(stdout, 'B\n');
+  });
 
 // Test preloading a single module works
 childProcess.exec(`"${nodeBinary}" ${preloadOption([fixtureA])} "${fixtureB}"`,
@@ -62,6 +78,32 @@ childProcess.exec(
   }
 );
 
+// Test that preload can be used with --frozen-intrinsics
+childProcess.exec(
+  `"${nodeBinary}" --frozen-intrinsics ${
+    preloadOption([fixtureE])
+  } ${
+    fixtureF
+  }`,
+  function(err, stdout) {
+    assert.ifError(err);
+    assert.strictEqual(stdout, 'smoosh\n');
+  }
+);
+childProcess.exec(
+  `"${
+    nodeBinary
+  }" --frozen-intrinsics ${
+    preloadOption([fixtureE])
+  } ${
+    fixtureG
+  } ${fixtureF}`,
+  function(err, stdout) {
+    assert.ifError(err);
+    assert.strictEqual(stdout, 'smoosh\n');
+  }
+);
+
 // Test that preload can be used with stdin
 const stdinProc = childProcess.spawn(
   nodeBinary,
@@ -86,7 +128,7 @@ const replProc = childProcess.spawn(
 );
 replProc.stdin.end('.exit\n');
 let replStdout = '';
-replProc.stdout.on('data', function(d) {
+replProc.stdout.on('data', (d) => {
   replStdout += d;
 });
 replProc.on('close', function(code) {
@@ -94,8 +136,9 @@ replProc.on('close', function(code) {
   const output = [
     'A',
     '> '
-  ].join('\n');
-  assert.strictEqual(replStdout, output);
+  ];
+  assert.ok(replStdout.startsWith(output[0]));
+  assert.ok(replStdout.endsWith(output[1]));
 });
 
 // Test that preload placement at other points in the cmdline
@@ -112,9 +155,8 @@ childProcess.exec(
 // Test that preload works with -i
 const interactive = childProcess.exec(
   `"${nodeBinary}" ${preloadOption([fixtureD])}-i`,
-  common.mustCall(function(err, stdout, stderr) {
-    assert.ifError(err);
-    assert.strictEqual(stdout, "> 'test'\n> ");
+  common.mustSucceed((stdout, stderr) => {
+    assert.ok(stdout.endsWith("> 'test'\n> "));
   })
 );
 
@@ -134,8 +176,7 @@ childProcess.exec(
 childProcess.exec(
   `"${nodeBinary}" ${preloadOption(['./printA.js'])} "${fixtureB}"`,
   { cwd: fixtures.fixturesDir },
-  common.mustCall(function(err, stdout, stderr) {
-    assert.ifError(err);
+  common.mustSucceed((stdout, stderr) => {
     assert.strictEqual(stdout, 'A\nB\n');
   })
 );
@@ -144,8 +185,7 @@ if (common.isWindows) {
   childProcess.exec(
     `"${nodeBinary}" ${preloadOption(['.\\printA.js'])} "${fixtureB}"`,
     { cwd: fixtures.fixturesDir },
-    common.mustCall(function(err, stdout, stderr) {
-      assert.ifError(err);
+    common.mustSucceed((stdout, stderr) => {
       assert.strictEqual(stdout, 'A\nB\n');
     })
   );
